/*
 * NgmRefWriter.cpp
 *
 *  Created on: May 4, 2011
 *      Author: philipp_
 */

#include "NgmRefWriter.h"

#include <iomanip>

NgmRefWriter::NgmRefWriter(char const * const fileName) :
	NgmWriter(sig, entryLength, fileName) {
	dataCompl.open((std::string(fileName) + ".bsc").c_str(), std::ios::out);
	dataCompl.exceptions(ofstream::failbit | ofstream::badbit);
}

NgmRefWriter::~NgmRefWriter() {
	dataCompl.close();
}

char complement(char const c) {
	switch (c) {
	case 'A':
		return 'T';
	case 'a':
		return 'T';
	case 'T':
		return 'A';
	case 't':
		return 'A';
	case 'G':
		return 'C';
	case 'g':
		return 'C';
	case 'C':
		return 'G';
	case 'c':
		return 'G';
	case 'N':
		return 'N';
	case 'n':
		return 'N';
	default:
		if (isalpha(c)) {
			if (islower(c)) {
				return 'N';
			} else {
				return 'N';
			}
		} else {
			std::string err("Invalid character: ");
			throw(err + c).c_str();
		}
	}
}

FilePosition NgmRefWriter::writeReference(char * sequence, int const length,
		int const padding) {
	FilePosition outputPosition = writeReference(data, sequence, length,
			padding, false);
	FilePosition outputPositionCheck = writeReference(dataCompl, sequence,
			length, padding, true);
	if (outputPosition != outputPositionCheck) {
		throw "Invalid output position in function convertRef!\n";
	}
	return outputPosition;
}

FilePosition NgmRefWriter::writeReference(ofstream & out, char * sequence,
		int const length, int const padding, bool reverse) {
	FilePosition outputPosition = out.tellp();

	toUpperCase(sequence, length);
	out << std::left << std::setw(padding) << std::setfill('x') << "";
	if (reverse) {
		for (int i = length - 1; i >= 0; --i) {
			out << std::setw(0) << complement(sequence[i]);
		}
	} else {
		out << sequence;
	}
	out << std::left << std::setw(padding) << std::setfill('x') << "";
	return outputPosition;
}

void NgmRefWriter::writeIndexEntryRef(FilePosition outputOffset,
		int outputLength, FilePosition nameOffset, int nameLength,
		bool isValid, bool isComplValid) {

	int flags = 0;
	if (isValid) {
		flags = flags | 1;
	}
	if (isComplValid) {
		flags = flags | 2;
	}

	idx.write((char*) &sequenceId, 4);
	idx.write((char*) &flags, 4);
	idx.write((char*) &outputOffset, 8);
	idx.write((char*) &nameOffset, 8);
	idx.write((char*) &outputLength, 4);
	idx.write((char*) &nameLength, 4);

	sequenceId += 1;
}
//
void NgmRefWriter::writeRefHeader(int const refCount, int const flags, int const padding) {
	WriteHeader(refCount, padding, flags, 0);
}
