#include "BAMWriter.h"

#ifdef _BAM

#include <string.h>

#include "api/BamWriter.h"
#include "api/SamHeader.h"
#include "api/BamAlignment.h"

#include "Config.h"
#include "SequenceProvider.h"
#include "NGM.h"

using namespace BamTools;

//using BamTools::BamWriter;
//using BamTools::SamHeader;
//Format: http://samtools.sourceforge.net/SAM1.pdf

static inline void rev(char * s);

BamWriter writer;

void BAMWriter::DoWriteProlog(FILE * output) {
	//TODO: check correct format


	SamHeader header;
	RefVector refs;

	header.Version = "1.0";
	header.SortOrder = "unsorted";

	for (int i = 0; i < SequenceProvider.GetRefCount(); ++i) {
		int len;
		char const * const refname = SequenceProvider.GetRefName(i, len);

		RefData ref(std::string(refname, len), SequenceProvider.GetRefLen(i));
		refs.push_back(ref);
		if (NGM.DualStrand())
			++i;
	}

	//			fprintf(output, "@PG\tID:ngm\tVN:%s\tCL:\"%s\"\n", "0.0.1",
	//					Config.GetString("cmdline"));


	//
	//		int ref = Config.GetInt("ref_mode");

	//		if (ref == -1)
	//			for (int i = 0; i < SequenceProvider.GetRefCount(); ++i) {
	//				fprintf(output, "@SQ\tSN:%.*s\tLN:%i\n", len,
	//						SequenceProvider.GetRefName(i, len),
	//						SequenceProvider.GetRefLen(i));
	//				if (NGM.DualStrand())
	//					++i;
	//			}
	//		else
	//			fprintf(output, "@SQ\tSN:%.*s\tLN:%i\n", len,
	//					SequenceProvider.GetRefName(ref * ((NGM.DualStrand()) ? 2
	//							: 1), len), SequenceProvider.GetRefLen(ref
	//							* ((NGM.DualStrand()) ? 2 : 1)));
	//
	//		if (Config.GetInt("append") != 1) {
	//			//TODO: add version

	//		}
	if (!writer.Open(file, header, refs)) {
		Log.Error("Could not open output BAM file");
		return;
	}

}

void BAMWriter::DoWriteRead(FILE * output, MappedRead const * const read) {
	static int const report_offset = 1;

	if(read->TLS()->Location.m_Location + report_offset >0){
		BamAlignment al;
		al.AlignmentFlag = 0;
		al.SetIsMapped(true);


		if(paired_end){
			//paired end mapping mode

			al.SetIsPaired (true);
			if( read->ReadId & 0x1){
				//if it is the second pair
				al.SetIsSecondMate(true);
				al.SetIsFirstMate(false);
			}else{
				//if it is the first pair
				al.SetIsSecondMate(false);
				al.SetIsFirstMate(true);
			}
			if(read->Paired!=0){
				al.SetIsProperPair(IsProperPaired(read));
				al.SetIsMateMapped((read->Paired->Identity > identity));
				al.SetIsMateReverseStrand((read->Paired->Strand == '-'));
			}
		}



		char const * readseq = SequenceProvider.GetQrySequence(read->ReadId);
		int readlen = SequenceProvider.GetQryLen(read->ReadId);
		if ((read->Strand == '-')) {
			char * tmp = new char[readlen + 1];
			memcpy(tmp, readseq, readlen + 1);
			rev(tmp);
			readseq = tmp;
			//TODO: also reverse quality values???

			al.SetIsReverseStrand(true);
		}



		int readnamelen;
		char const * readname = SequenceProvider.GetQryName(read->ReadId,
				readnamelen);
		al.Name = std::string(readname, readnamelen);
		al.Length = readlen;
		al.MapQuality = 255;
		al.Position = read->TLS()->Location.m_Location+ report_offset;
		std::string seq(readseq, readlen);
		al.QueryBases = seq;
		al.RefID = read->Scores[read->TopScore]->Location.m_RefId / 2;

		//TODO: Add support for paired end
		//	fprintf(output, "*\t"); //Ref. name of the mate/next fragment
		//	fprintf(output, "0\t"); //Position of the mate/next fragment
		//	fprintf(output, "0\t"); //observed Template LENgth

		int length = atoi(read->Buffer1);
		int i = 0;
		while (read->Buffer1[i] != '\0') {
			if (read->Buffer1[i] < '0' || read->Buffer1[i] > '9') {
				al.CigarData.push_back(CigarOp(read->Buffer1[i], length));
				length = atoi(read->Buffer1 + i + 1);
			}
			i += 1;
		}

		int qltylen = 0;
		char const * qltystr = SequenceProvider.GetQualityString(read->ReadId,
				qltylen);
		if (qltylen > 0 && qltylen == readlen) {
			al.Qualities = std::string(qltystr, qltylen);
		} else {
			al.Qualities = std::string(readlen, ':');
		}

		//	//Optional fields
		al.AddTag("AS", "i", (int) read->Scores[read->TopScore]->Score.f);
		al.AddTag("MD", "Z", std::string(read->Buffer2));
		al.AddTag("X0", "i", (int) read->EqualScoringCount);
		al.AddTag("XI", "f", read->Identity);

		if ((read->Strand == '-')) {
			delete[] readseq;
			readseq = 0;
		}

		writer.SaveAlignment(al);
	}
}

void BAMWriter::DoWriteEpilog(FILE * fp) {
	writer.Close();
}

//TODO: Use functions from Output.cpp
static inline char cpl(char c) {
	if (c == 'A')
		return 'T';
	else if (c == 'T')
		return 'A';
	else if (c == 'C')
		return 'G';
	else if (c == 'G')
		return 'C';
	else
		return c;
}

// swaps two bases and complements them
static inline void rc(char & c1, char & c2) {
	char x = c1;
	c1 = cpl(c2);
	c2 = cpl(x);
}

// Reverse-complements a 0-terminated string in place
static inline void rev(char * s) {
	char * end = s + strlen(s) - 1;

	while (s < end)
		rc(*s++, *end--);

	if (s == end)
		*s = cpl(*s);
}

bool  BAMWriter::IsProperPaired(MappedRead const * const read){

	bool same_chr= (read->Scores[read->TopScore]->Location.m_RefId == read->Paired->Scores[read->Paired->TopScore]->Location.m_RefId);

	int dist = abs(read->TLS()->Location.m_Location - read->Paired->TLS()->Location.m_Location);

	return true;// ((read->Identity > identity && read->Paired->Identity > identity)
	//&&(read->Paired->Strand != read->Strand) && same_chr
	//&& (dist<max_dist && dist>min_dist) ) ;

}



#endif
