#include "CIGARWriter.h"

#include <string.h>

#include "Config.h"
#include "SequenceProvider.h"
#include "NGM.h"

//Format: http://samtools.sourceforge.net/SAM1.pdf

static inline void rev(char * s);

void CIGARWriter::DoWriteProlog(FILE * output) {
	//TODO: check correct format
	if (Config.GetInt("append") != 1) {
		fprintf(output, "@HD\tVN:1.0\tSO:unsorted\n");

		int ref = Config.GetInt("ref_mode");
		int len;
		if (ref == -1)
			for (int i = 0; i < SequenceProvider.GetRefCount(); ++i) {
				fprintf(output, "@SQ\tSN:%.*s\tLN:%i\n", len,
						SequenceProvider.GetRefName(i, len),
						SequenceProvider.GetRefLen(i));
				if (NGM.DualStrand())
					++i;
			}
		else
			fprintf(
					output,
					"@SQ\tSN:%.*s\tLN:%i\n",
					len,
					SequenceProvider.GetRefName(
							ref * ((NGM.DualStrand()) ? 2 : 1), len),
					SequenceProvider.GetRefLen(
							ref * ((NGM.DualStrand()) ? 2 : 1)));

		if (Config.GetInt("append") != 1) {
			//TODO: add version
			fprintf(output, "@PG\tID:ngm\tVN:%s\tCL:\"%s\"\n", "0.0.1",
					Config.GetString("cmdline"));
		}
	}

}

void CIGARWriter::DoWriteRead(FILE * output, MappedRead const * const read) {
	static int const report_offset = 1;

	if(read->TLS()->Location.m_Location + report_offset >0){
		int readnamelen;
		int refnamelen;
		int qltylen;
		int flags = 0;
		char const * readname = SequenceProvider.GetQryName(read->ReadId,
				readnamelen);
		char const * qltystr = SequenceProvider.GetQualityString(read->ReadId,
				qltylen);
		char const * readseq = SequenceProvider.GetQrySequence(read->ReadId);
		int readlen = SequenceProvider.GetQryLen(read->ReadId);


		if(paired_end){
			//paired end mapping mode
			flags |= 0x1;

			if( read->ReadId & 0x1){
				//if it is the second pair
				flags |= 0x80;
			}else{
				//if it is the first pair
				flags |= 0x40;
			}

			if(read->Paired!=0){
				Log.Warning("Is paired end.");
				if(IsProperPaired(read)){
					//read and its pair is mapped
					flags |= 0x2;
				}

				if(read->Paired->Identity > identity){
					flags |= 0x8;
				}
				if(read->Paired->Strand == '-'){
					//if the pair is mapped on - or + stran
					flags |= 0x20;
				}

			}
		}



		if(read->Identity < identity){
			//read is not mapped
			flags |= 0x4;
		}

		//TODO Check if we can compute mapping quality

		if ((read->Strand == '-')) {
			char * tmp = new char[readlen + 1];
			memcpy(tmp, readseq, readlen + 1);
			rev(tmp);
			readseq = tmp;
			//TODO: also reverse quality values???
			flags |= 0x10;
		}


		char const * refname = SequenceProvider.GetRefName(
				read->Scores[read->TopScore]->Location.m_RefId, refnamelen);

		//mandatory fields
		fprintf(output, "%.*s\t", readnamelen, readname);
		fprintf(output, "%d\t", flags);
		fprintf(output, "%.*s\t", refnamelen, refname);
		fprintf(output, "%i\t", read->TLS()->Location.m_Location + report_offset);
		fprintf(output, "255\t");
		fprintf(output, "%s\t", read->Buffer1);
		//TODO: Add support for paired end
		fprintf(output, "*\t"); //Ref. name of the mate/next fragment
		fprintf(output, "0\t"); //Position of the mate/next fragment
		fprintf(output, "0\t"); //observed Template LENgth
		fprintf(output, "%s\t", readseq);
		if (qltylen > 0 && qltylen == readlen) {
			fprintf(output, "%.*s\t", qltylen, qltystr);
		} else {
			fprintf(output, "*\t");
		}

		//Optional fields
		fprintf(output, "AS:i:%d", (int) read->Scores[read->TopScore]->Score.f);
		fprintf(output, "\tMD:Z:%s", read->Buffer2);
		fprintf(output, "\tX0:i:%d", read->EqualScoringCount);
		fprintf(output, "\tXI:f:%f", read->Identity);

		fprintf(output, "\n");

		if ((read->Strand == '-')) {
			delete[] readseq;
			readseq = 0;
		}
	}
}

void CIGARWriter::DoWriteEpilog(FILE * fp) {

}

//TODO: Use functions from Output.cpp
static inline char cpl(char c) {
	if (c == 'A')
		return 'T';
	else if (c == 'T')
		return 'A';
	else if (c == 'C')
		return 'G';
	else if (c == 'G')
		return 'C';
	else
		return c;
}

// swaps two bases and complements them
static inline void rc(char & c1, char & c2) {
	char x = c1;
	c1 = cpl(c2);
	c2 = cpl(x);
}

// Reverse-complements a 0-terminated string in place
static inline void rev(char * s) {
	char * end = s + strlen(s) - 1;

	while (s < end)
		rc(*s++, *end--);

	if (s == end)
		*s = cpl(*s);
}

bool  CIGARWriter::IsProperPaired(MappedRead const * const read){

	bool same_chr= (read->Scores[read->TopScore]->Location.m_RefId == read->Paired->Scores[read->Paired->TopScore]->Location.m_RefId);

	int dist = abs(read->TLS()->Location.m_Location - read->Paired->TLS()->Location.m_Location);

	return true;// ((read->Identity > identity && read->Paired->Identity > identity)
	//&&(read->Paired->Strand != read->Strand) && same_chr
	//&& (dist<max_dist && dist>min_dist) ) ;

}
