#ifndef __CIGARWRITER_H__
#define __CIGARWRITER_H__

#include "GenericReadWriter.h"

class CIGARWriter: public GenericReadWriter {
	public:
		CIGARWriter(char const * const filename) :
			GenericReadWriter(filename) {
			if (Config.Exists("paired")){

				paired_end = (Config.GetInt("paired", 0, 1) == 1);
				if (Config.Exists("pair_min_distance")&&Config.Exists("pair_max_distance")){
					min_dist= Config.GetInt("pair_min_distance");
					max_dist=Config.GetInt("pair_max_distance");
				}else{
					Log.Warning("No paired end distance is set.");
					min_dist=0;
					max_dist=999999999;
				}
			}
			identity=0;
			if (Config.Exists("identity_tresh")){
				identity=Config.GetFloat("identity_tresh",0,100);
				identity=identity/100.0f;
			}

		}

	protected:
		virtual void DoWriteProlog(FILE * fp);
		virtual void DoWriteRead(FILE * fp, MappedRead const * const read);
		virtual void DoWriteEpilog(FILE * fp);
	private:

		bool IsProperPaired(MappedRead const * const read);
		bool paired_end;
		int min_dist;
		int max_dist;
		float identity;
};

#endif
