#include "CS.h"
#include "NGM.h"

#include "CSCache.h"

#include "Timing.h"
#include <memory.h>
#include <stdlib.h>

#include "Debug.h"
#include "SW.h"

#undef module_name
#define module_name "CS"

#define CallMemberFn(obj,pFn)  ((obj).*(pFn))

volatile int CS::s_ThreadCount = 0;

// Default-Werte
int RefEntry::MaxRefsPerEntry = 100;	// Anzahl Positionen, die pro RefEntry-Knoten gespeichert werden

int const cBatchSize = 1000;
int const cThreshhold = 4;
int const cSrchTableLen = 9999991;
int const cREThresh = 1000;
float const cOverflowThresh = 0.1f;
float const cOverflowLimit = 10;
uint const cPrefixBaseSkip = 0;

uint CS::prefixBasecount = 10;
uint CS::prefixBits = prefixBasecount*2;
ulong CS::prefixMask = ((ulong)1 << prefixBits) - 1;

int const stageTableLen[8] = { 32749, 65521, 131071, 262139, 524287, 1048573, 2097143, 4194301 };

inline int min(int a, int b)
{
	return (a < b) ? a : b;
}

// A->0 C->1 T->2 G->3
inline char encode(char c)
{
	return (c >> 1) & 3;
}

// Iteriert ueber jeden Praefix in sequence und fuehrt fuer diesen die Funktion func aus
void CS::PrefixIteration(char const * sequence, uint length, PrefixIterationFn func, void* data, uint prefixskip, uint offset)
{
	if (length < prefixBasecount)
		return;

	if (*sequence == 'N')
	{
		uint n_skip = 1;
		while (*(sequence + n_skip) == 'N')
			++n_skip;

		sequence += n_skip;

		if (n_skip >= (length-prefixBasecount))
			return;
		length -= n_skip;
		offset += n_skip;
	}

	ulong prefix = 0;
	for (uint i = 0; i < prefixBasecount-1; ++i)
	{
		char c = *(sequence+i);
		if (c == 'N')
		{
			PrefixIteration(sequence+i+1, length-i-1, func, data, prefixskip, offset+i+1);
			return;
		}

		prefix = prefix << 2;
		char cx = encode(c);
		prefix |= cx;
	}

	uint skipcount = prefixskip;
	for (uint i = prefixBasecount-1; i < length; ++i)
	{
		char c = *(sequence+i);
		if (c == 'N')
		{
			PrefixIteration(sequence+i+1, length-i-1, func, data, prefixskip, offset+i+1);
			return;
		}

		prefix = prefix << 2;
		char cx = encode(*(sequence+i));
		prefix |= cx;
		prefix &= prefixMask;

		if (skipcount == prefixskip)
		{
			func(prefix, offset+i+1-prefixBasecount, data);
			skipcount = 0;
		}
		else
		{
			++skipcount;
		}
	}
}

// mutate T (0x2) -> C (0x1)
ulong const mutateFrom = 0x2; 
ulong const mutateTo = 0x1;

void CS::PrefixMutateSearch(ulong prefix, int pos, void* data)
{
	static int const cMutationLocLimit = Config.Exists("cs_mutationlimit") ? Config.GetInt("cs_mutationlimit") : 6;
	ulong const mask = 0x3;

	int mutationLocs = 0;
	for (int i = 0; i < (int)prefixBasecount; ++i)
	{
	  ulong base = mask & (prefix >> (i * 2));
	  if (base == mutateFrom)
		++mutationLocs;
	}

	if (mutationLocs <= cMutationLocLimit)
		PrefixMutateSearchEx(prefix, pos, data);
}

void CS::PrefixMutateSearchEx(ulong prefix, int pos, void* data, int mpos)
{
	PrefixSearch(prefix, pos, data);

	ulong const mask = 0x3;
	for (int i = mpos; i < (int)prefixBasecount; ++i)
	{
		ulong cur = mask & (prefix >> (i * 2));

		if (cur == mutateFrom)
		{
			ulong p1 = (prefix & ~(mask << (i * 2)));
			ulong p2 = (mutateTo << (i*2));
			PrefixMutateSearchEx( p1 | p2, pos, data, i+1);
		}
	}
}

void CS::PrefixSearch(ulong prefix, int pos, void* data)
{
	CS * cs = (CS*)data;

	RefEntry const * cur = cs->m_RefProvider->GetRefEntry(prefix);	// Liefert eine liste aller Vorkommen dieses Praefixes in der Referenz
	if (cur != 0)
	{
		if (cur->refTotal > cs->m_REThresh)
		{
			cur = 0;
		}
		else
		{
			++cs->m_TotalPrefixes;
		}
	}

	while (cur != 0)
	{
		//Log.Green("Qry Seq %i - Prefix 0x%x got %i locs (sum %i)", cs->m_CurrentSeq, prefix, cur->refTotal, locTotal);
		cs->m_TotalLocs += cur->refCount;

		for (int i = 0; i < cur->refCount; ++i)
		{
			SequenceLocation curLoc = cur->ref[i];
			curLoc.m_Location = cs->GetBin(curLoc.m_Location-pos);	// position offset

			CallMemberFn(*cs, cs->AddLocation)(curLoc);
		}

		cur = cur->nextEntry;
	}
}

void CS::AddLocationStd(SequenceLocation const & loc)
{
	int hpo = Hash(loc.m_Location);
	//       intelligentere overflow-detektion: irgendwas wie aktueller pr�fix �ber 50% des reads
	//       und f�llstand der tabelle mit pr�fixposition vergleichen -> wenn �ber einem threshhold,
	//       sofort abbrechen und auf fallback mode gehen
	int hpoc = m_SrchTableLen;
	while ( (rTable[hpo].Location.m_Location != cInvalidLocation) && !(rTable[hpo].Location.m_Location == loc.m_Location && rTable[hpo].Location.m_RefId == loc.m_RefId))
	{
		++hpo;
		if (hpo >= m_SrchTableLen)
			hpo = 0;
		if (--hpoc == 0)
		{
			throw 1;
		}
	}

	if (rTable[hpo].Location.m_Location == cInvalidLocation)
	{
		++m_UsedSlots;
		rTable[hpo].Location = loc;
	}

	if (++rTable[hpo].Score.i == m_Threshhold)
	{
		++m_Candidates;
	}
}

void CS::AddLocationFallback(SequenceLocation const & loc)
{
	if ( (++iTable[loc]) == m_Threshhold )
	{
		++m_Candidates;
	}
}

int CS::CollectResultsStd(MappedRead * read)
{
	Log.Verbose("Qry #%i got %i results", m_CurrentSeq, m_Candidates);

	if (m_Candidates == 0)
	{
		for (int i=0; i < m_SrchTableLen; ++i)
		{
			rTable[i] = sInvalidLocationScore;
		}
		m_UsedSlots = 0;
		return 1;
	}

	int n = 0;
	for (int i = 0; i < m_SrchTableLen; ++i)
	{
		if (rTable[i].Location.m_Location != cInvalidLocation)
		{
			if (rTable[i].Score.i >= m_Threshhold)
			{
				rTable[i].Location.m_Location = ResolveBin(rTable[i].Location.m_Location);
				read->AddScore(rTable[i]);
				//Log.Green("Adding Location <%i, %i> with Score %i", rTable[i].Location.m_Location, rTable[i].Location.m_RefId, rTable[i].Score.i);
				++n;
			}
			rTable[i] = sInvalidLocationScore;
			--m_UsedSlots;
		}
	}

	if (n != m_Candidates)
	{
		Log.Error("Candidate list size mismatch on read %i (n = %i, m_C = %i)", m_CurrentSeq, n, m_Candidates);
	}

	return 0;
}

int CS::CollectResultsFallback(MappedRead * read)
{
	int n = 0;
	for (std::map<SequenceLocation, int>::iterator itr = iTable.begin(); itr != iTable.end(); itr++)
	{
		if (itr->second >= m_Threshhold)
		{
			LocationScore score (itr->first, itr->second, read);

			score.Location.m_Location = ResolveBin(score.Location.m_Location);
			read->AddScore( score );
			//Log.Message("Adding location %i (score %i)", itr->first.m_Location, itr->second);
			++n;
		}
	}
	iTable.clear();
	if (n != m_Candidates)
	{
		Log.Error("Candidate list size mismatch on read %i (n = %i, m_C = %i)", m_CurrentSeq, n, m_Candidates);
	}
	return 0;
}

void CS::SendToBuffer(MappedRead * read)
{
	if (read == 0)
		return;

	int count = read->Scores.size() - read->Calculated;

	Log.Verbose("Sending %i candidates for read %i to buffer", count, read->ReadId);

	if (count == 0)
	{
		if (NGM.LastPartition())
		{
			SW::SendToPostprocessing(read);
		}
		else
		{
			NGM.BufferRead(read);
			++m_WrittenReads;
		}
	}
	else
	{
		if (count < NGM.bCSSW.Capacity())
		{
			NGM.bCSSW.Write(&read->Scores[read->Calculated], count);
			++m_WrittenReads;
		}
		else
		{
			Log.Warning("Read %i discarded due to buffer overflow", read->ReadId);
			++m_DiscardedReads;
		}
	}
}

void CS::Cleanup()
{
}

void CS::RunBatch()
{
	PrefixIterationFn pFunc = &CS::PrefixSearch;
	if (m_EnableBS)
		pFunc = &CS::PrefixMutateSearch;

	for (size_t i = 0; i < m_CurrentBatch.size(); ++i)
	{
		m_CurrentSeq = m_CurrentBatch[i]->ReadId;
		++m_ProcessedReads;
		bool fallback = m_Fallback;

		if (m_CSThreadID < NGMStats::cMaxThreadStats)
			NGM.Stats->CS[m_CSThreadID].CurrentRead = m_CurrentSeq;

		char const * const qrySeq = SequenceProvider.GetQrySequence(m_CurrentSeq);
		int qryLen = strlen(qrySeq);

		if (!fallback)
		{
			try
			{
				m_Candidates = 0;
				m_TotalLocs = 0;
				m_TotalPrefixes = 0;
				if (m_UsedSlots > 0)
				{
					Log.Warning("Unclean CS Hash table @ read %i (Slots used: %i)", m_CurrentSeq, m_UsedSlots);
				}

				AddLocation = &CS::AddLocationStd;
				PrefixIteration(qrySeq, qryLen, pFunc, this, m_PrefixBaseSkip);
				CollectResultsStd(m_CurrentBatch[i]);
			}
			catch (int overflow)
			{
				Log.Message("Interrupt on read %i: %i candidates in %i slots filled by %i prefixes containing %i locs",
						m_CurrentSeq,
						m_Candidates,
						m_UsedSlots,
						m_TotalPrefixes,
						m_TotalLocs);

				++m_Overflows;
				CheckFallback();

				// Clear rTable
				for (int i = 0; i < m_SrchTableLen; ++i)
					rTable[i] = sInvalidLocationScore;
				m_UsedSlots = 0;

				// Initiate fallback search on iTable
				fallback = true;
			}
		}
		if (fallback)
		{
			m_Candidates = 0;

			AddLocation = &CS::AddLocationFallback;
			PrefixIteration(qrySeq, qryLen, pFunc, this, m_PrefixBaseSkip);
			CollectResultsFallback(m_CurrentBatch[i]);
		}

		SendToBuffer(m_CurrentBatch[i]);
	}
}

void CS::DoRun()
{
	Log.Message("Launching CS Thread %i (NGM Thread %i)", m_CSThreadID, m_TID);
	rTable = new LocationScore[m_SrchTableLen];
	for (int i = 0; i < m_SrchTableLen; ++i)
	{
		rTable[i] = sInvalidLocationScore;
	}

	if (Config.Exists("cs_enablebs"))
		m_EnableBS = (Config.GetInt("cs_enablebs", 0, 1) == 1);

	m_RefProvider = 0;

	Timer tmr;

	while ( NGM.ThreadActive(m_TID, GetStage()) && ((m_CurrentBatch = NGM.GetNextReadBatch(m_BatchSize)), (m_CurrentBatch.size() > 0)) )
	{
		Log.Verbose("CS Thread %i got batch (len %i)", m_TID, m_CurrentBatch.size());
		tmr.ST();

		m_RefProvider = NGM.GetRefProvider(m_TID);

		NGM.bCSSW.Register();
		RunBatch();
		NGM.bCSSW.Release();

		NGM.ReleaseRefProvider(m_TID);

		Log.Message("CS Thread %i finished batch (len %i) (elapsed: %.3fs)", m_TID, m_CurrentBatch.size(), tmr.ET());
	}

	Log.Message("CS Thread %i finished (%i reads processed, %i reads written, %i reads discarded)", m_TID, m_ProcessedReads, m_WrittenReads, m_DiscardedReads);
}

void CS::Init()
{
	prefixBasecount = Config.GetInt("cs_prefixsize", 4, 32);
	prefixBits = prefixBasecount*2;
	prefixMask = ((ulong)1 << prefixBits) - 1;
}

CS::CS(bool useBuffer) : m_CSThreadID( (useBuffer) ? (AtomicInc( &s_ThreadCount ) - 1) : -1 ),
			m_BatchSize( (Config.Exists("cs_batchsize")) ? Config.GetInt("cs_batchsize") : cBatchSize ),
			m_ProcessedReads( 0 ),
			m_WrittenReads( 0 ),
			m_DiscardedReads( 0 ),
			m_EnableBS( false ),
			m_UsedSlots( 0 ),
			m_TotalLocs( 0 ),
			m_TotalPrefixes( 0 ),
			m_Overflows( 0 ),
			m_SrchTableLen( (Config.Exists("cs_searchTableLen")) ? Config.GetInt("cs_searchTableLen") : cSrchTableLen ),
			m_Threshhold( (Config.Exists("cs_threshhold")) ? Config.GetInt("cs_threshhold") : cThreshhold ),
			m_REThresh( (Config.Exists("cs_rethresh")) ? Config.GetInt("cs_rethresh") : cREThresh ),
			m_PrefixBaseSkip( (Config.Exists("cs_prefixskip")) ? Config.GetInt("cs_prefixskip", 0, -1) : cPrefixBaseSkip ),
			m_Fallback( (m_SrchTableLen <= 0) ),  // cTableLen <= 0 means always use fallback
			AddLocation( 0 )
	{
		rTable = 0;
	}

CS::~CS()
{
	if (rTable != 0)
		delete[] rTable;
}


void CS::CheckRTable()
{
	for (int i = 0; i < m_SrchTableLen; ++i)
	{
		if (rTable[i].Location.m_Location != cInvalidLocation)
		{
			Log.Warning("Unclean rTable (qry = %i, elem = %i): loc = %i", m_CurrentSeq, i, rTable[i].Location.m_Location);
			rTable[i] = sInvalidLocationScore;
		}
	}
}

void CS::CheckFallback()
{
	static float sOverflowThresh = (Config.Exists("cs_overflowthresh")) ? Config.GetFloat("cs_overflowthresh", 0.0f, 1.0f) : cOverflowThresh;
	static int sOverflowLimit = (Config.Exists("cs_overflowlimit")) ? Config.GetInt("cs_overflowlimit") : cOverflowLimit;

	if (sOverflowLimit <= 0)
	{
		m_Fallback = true;
	}

	if (m_Overflows < sOverflowLimit)
		return;

	float overflowRatio = (float)m_Overflows / (float)m_ProcessedReads;
	if ( overflowRatio > sOverflowThresh )
	{
		m_Fallback = true;
		Log.Warning("Permanently switching to fallback search (%i out of %i reads overflowed)", m_Overflows, m_ProcessedReads);
	}
}

inline int CS::Hash(int n)
{
	return abs(n) % m_SrchTableLen;
}
