#include "FFormatWriter.h"

#include "Config.h"
#include "SequenceProvider.h"
#include "NGM.h"

#include <string>
std::string add_timestamp(std::string);

void FFormatWriter::DoWriteProlog(FILE * output)
{

	if (Config.Exists("cs_enablebs")){
		m_EnableBS = (Config.GetInt("cs_enablebs", 0, 1) == 1);

	}


	fprintf(output, "%s", add_timestamp("NGM Output (generated %s)\n\n").c_str());

	int ref = Config.GetInt("ref_mode");
	int cnt = SequenceProvider.GetRefCount() >> (NGM.DualStrand() ? 1 : 0);
	fprintf(output, "Reference sequence/chromosome count (used for search/total in file): %i / %i\n", (ref == -1) ? cnt : 1, cnt);
	int len;
	if (ref == -1)
		for (int i = 0; i < SequenceProvider.GetRefCount(); ++i)
		{
			fprintf(output, "> %.*s\n", len, SequenceProvider.GetRefName(i, len));
			fprintf(output, "  Length: %i\n", SequenceProvider.GetRefLen(i));

			if (NGM.DualStrand())
				++i;
		}
	else
		fprintf(output, "> %.*s\n", len, SequenceProvider.GetRefName(ref * ((NGM.DualStrand()) ? 2 : 1), len));
	fprintf(output, "\n");
}

void FFormatWriter::DoWriteRead(FILE * output, MappedRead const * const read)
{
	static int const report_offset = Config.GetInt("report_offset");

	int readnamelen; int refnamelen; int qltylen;
	char const * readname = SequenceProvider.GetQryName(read->ReadId, readnamelen);
	char const * refname = SequenceProvider.GetRefName(read->Scores[read->TopScore]->Location.m_RefId, refnamelen);
	int refnamelenmod = refnamelen;
	if (refnamelenmod >= 4) // Wenn refname mehr als 3 Zeichen lang ist, wird beim ersten Leerzeichen abgebrochen
	{
		refnamelenmod = 4;
		while (refnamelenmod < refnamelen && *(refname+refnamelenmod) != ' ')
			++refnamelenmod;
	}
	char const * qltystr = SequenceProvider.GetQualityString(read->ReadId, qltylen);


	if(m_EnableBS &&  read->ReadId & 0x1){
		if(read->ReadId & 0x1){
			//read from file 2:
			fprintf(output, "ID: %i.%i\tRead: %.*s\tDir: %c%c\tPos: %i\tChr: %.*s\tScore: %.1f\n",
				read->ReadId,
				read->EqualScoringID,
				readnamelen,
				readname,
				'-', read->Strand,
				read->TLS()->Location.m_Location + report_offset,
				refnamelenmod,
				refname,
				read->Scores[read->TopScore]->Score.f
			);
		}else {
			fprintf(output, "ID: %i.%i\tRead: %.*s\tDir: %c%c\tPos: %i\tChr: %.*s\tScore: %.1f\n",
				read->ReadId,
				read->EqualScoringID,
				readnamelen,
				readname,
				'+', read->Strand,
				read->TLS()->Location.m_Location + report_offset,
				refnamelenmod,
				refname,
				read->Scores[read->TopScore]->Score.f
			);
		}
	}else{
		fprintf(output, "ID: %i.%i\tRead: %.*s\tDir: %c\tPos: %i\tChr: %.*s\tScore: %.1f\n",
			read->ReadId,
			read->EqualScoringID,
			readnamelen,
			readname,
			read->Strand,
			read->TLS()->Location.m_Location + report_offset,
			refnamelenmod,
			refname,
			read->Scores[read->TopScore]->Score.f
		);


	}
		fprintf(output, "times: %i\tidentity: %f\tQ_start: %i\tQ_end: %i\n",
		read->EqualScoringCount,
		read->Identity,
		read->QStart, 
		read->QEnd
		);

	// �berspringe Leerzeichen am Anfang
	int space_skip = 0;
	while (read->Buffer1[space_skip] == ' ')
		++space_skip;

	fprintf(output, "%s\n%s\n",
		read->Buffer1+space_skip, 
		read->Buffer2+space_skip
		);

	// Output Quality-String (+gaps)
	if (qltylen > 0)
	{
		// TODO: QStart, QEnd, Strand reincodieren
		int pos = 0;
		int qpos = 0;
		while (read->Buffer2[space_skip+pos] != 0)
		{
			if (read->Buffer2[space_skip+pos] == '-')
				fprintf(output, "-");
			else
				fprintf(output, "%c", *(qltystr+qpos++));
			++pos;
		}
	}

	fprintf(output, "\n\n");
}

void FFormatWriter::DoWriteEpilog(FILE * output)
{
}

