#ifndef __MAPPEDREAD_H__
#define __MAPPEDREAD_H__

#include "LocationScore.h"

#include <memory.h>
#include <vector>
#include <algorithm>
#include "Types.h"
#include "SequenceProvider.h"
#include "ILog.h"
#include "ReadStatus.h"

using NGMNames::ReadStatus;
// A read with a list of scores fitting the initial predicate
// (i.e., each score is greater than cs_threshhold)
struct MappedRead
{
private:

public:
	int const ReadId;
	volatile int Calculated; // Amount of Scores updated by SW
	int TopScore;			// Position of TopScore
	volatile int Lock;		// sync for paired read selection
	int EqualScoringID;		// Number of equal scoring result for this read
	uint EqualScoringCount;	// Total number of equal scoring results
	std::vector<LocationScore*> Scores;
	MappedRead * Paired;	// Read pair
	uint Status;
	float Identity;
	char Strand;
	int QStart;
	int QEnd;

	char * Seq;
	
	char * Buffer1;
	char * Buffer2;

#ifdef INSTANCE_COUNTING
	static volatile int sInstanceCount;
#endif

	MappedRead(int const readid);
	~MappedRead();

	// Saves score to this Read and returns a pointer to the saved object
	LocationScore * AddScore(LocationScore const & score);
	LocationScore * TLS() const;
	int nScores() const;
	void TopN(uint n);

	void AllocBuffers();

	inline void SetFlag(ReadStatus const flag)
	{
		Status |= flag;
	}
	inline bool HasFlag(ReadStatus const flag)
	{
		return (Status & flag) != 0;
	}
private:
	void Unique();

	static bool SortPred(LocationScore * lhs, LocationScore * rhs)
	{
		return lhs->Score.f > rhs->Score.f;
	}
	static bool UniquePred(LocationScore * lhs, LocationScore * rhs)
	{
		if (lhs == 0 || rhs == 0)
			return false;
		return (lhs->Location.m_Location == rhs->Location.m_Location) && (lhs->Location.m_RefId == rhs->Location.m_RefId);
	}
	static bool IsZero(LocationScore * arg)
	{
		return (arg == 0);
	}
};


#endif
