#include <stdlib.h>
#include <stdio.h>

#ifdef _WIN32
#include <crtdbg.h>
#endif

#include "NGM.h"

#include "SW.h"
#include "CS.h"
#include "Output.h"


#include "Timing.h"

#include "Debug.h"

#undef module_name
#define module_name "MAIN"

void CheckOutput()
{
	if (!(Config.Exists("overwrite") && (Config.GetInt("overwrite") == 1)))
	{
		if (FileExists(Config.GetString("output")))
		{
			Log.Error("Output file already exists and overwrite set to false.");
			Fatal();
		}
	}
}

#ifdef _DEBUG
	bool cDebug = true;
#else
	bool cDebug = false;
#endif

IConfig * _config = 0;

int main(int argc, char * argv[])
{
	try
	{
		char const * arch[] = { "x86", "x64" };
		char const * build = (cDebug) ? " (DEBUG)" : "";
		Log.Message("Startup : %s%s (build %s %s)", arch[sizeof(void*) / 4 - 1], build, __DATE__, __TIME__);

		Timer tmr;
		tmr.ST();

		_NGM::AppName = argv[0];

		InitPlatform();

		// Initialization:
		_config = new _Config(argc, argv);	// Parses command line & parameter file
		_Log::Init();				// Inits logging to file

		CheckOutput();

		if (Config.Exists("master_cpu"))
			NGMSetThreadAffinity(0, Config.GetInt("master_cpu"));

		NGM;						// Init Core
		SequenceProvider.Init();	// Prepares input data
		NGM.GeneratePartitions();

		((_Config*)_config)->Override("qry_max_len", SequenceProvider.GetReadMaxLen());

		Log.Message("Core initialization complete");

		NGM.StartThreads();

		NGM.MainLoop();

		Log.Message("Done (%i reads mapped (%.2f%%), %i reads not mapped, %i reads written)(elapsed: %fs)", NGM.GetMappedReadCount(), (float)NGM.GetMappedReadCount() * 100.0f / (float)(std::max(1, NGM.GetMappedReadCount()+NGM.GetUnmappedReadCount())),NGM.GetUnmappedReadCount(), NGM.GetWrittenReadCount(), tmr.ET());
	}
	catch (...)
	{
		Log.Error("Unhandled exception in control thread");
	}

	CS::Cleanup();
	_SequenceProvider::Cleanup();
	delete _config;
	delete _NGM::pInstance;
	_Log::Cleanup();


	CleanupPlatform();

	return 0;
}

void Help()
{
	char const * const help_msg = "\
./NGM [parameter]\n\
\n\
Parameters:\n\
\n\
-p filename	Parameterfile (alternative: \"-c parameterfile\")\n\
		This Parameter is required.\n\
			\n\
-h		Prints help and aborts program\n\
-d		Creates a default config file and aborts program\n\
\n\
\n\
-i filename	Readfile	(alternative: \"-c:qry filename\")\n\
-r filename	Referenzfile	(alternative: \"-c:ref filename\")\n\
-o filename	Outputfile	(alternative: \"-c:output filename\")\n\
-l filename	Logfile		(alternative: \"-c:logfile filename -c:log 1\")\n\
\n\
-c:name value	Sets config entry \"name\" to \"value\" (example: -c:score_match 2)\n\n";

	printf(help_msg);
	exit(0);
}

// actually platform specific.../care
ulong const FileSize(char const * const filename)
{
	FILE * fp = fopen(filename, "rb");
	if (fp == 0)
	{
		Log.Warning("Tried to get size of nonexistant file %s", filename);
		return 0;
	}

	if (fseek (fp, 0, SEEK_END) != 0)
		return 0;

	ulong end = ftell(fp);
	fclose(fp);
	return end;
}
