#ifndef __OUTPUT_H__
#define __OUTPUT_H__

#include "NGMTask.h"
#include "SW.h"
#include "GenericReadWriter.h"

#include <map>
#include <list>
#include <algorithm>

#undef module_name
#define module_name "OUT"

typedef void (*pfDelete)(void*);

char Strand(MappedRead * read);

class Output : public NGMTask
{
public:
	Output(char const * const filename) :
		m_Filename( filename )
	{
	}

	void DoRun();
	int GetStage() const
	{
		return 4;
	}
	inline char const * GetName() const { return "Output"; }

private:
	char const * const m_Filename;

	GenericReadWriter * m_Writer;

	std::map< int, std::list<MappedRead*> > m_EqualScoringBuffer;

	void SaveRead(MappedRead * read)
	{
		NGM.AquireOutputLock();
		m_Writer->WriteRead(read);
		NGM.AddWrittenRead(read->ReadId);
		NGM.ReleaseOutputLock();
	}

	static bool EqualScoringSortPredicate(MappedRead * lhs, MappedRead * rhs)
	{
		return (lhs->Strand == rhs->Strand) 
			? lhs->TLS()->Location.m_Location < rhs->TLS()->Location.m_Location
			: lhs->Strand < rhs->Strand;
	}

	static bool EqualScoringEquivalentPredicate(MappedRead * lhs, MappedRead * rhs)
	{
		return (lhs->TLS()->Location.m_RefId == rhs->TLS()->Location.m_RefId)
			&& (lhs->TLS()->Location.m_Location == rhs->TLS()->Location.m_Location);
	}
};

#endif
