#include "PrefixTable.h"
#include "NGM.h"
#include "CS.h"

#include "Timing.h"

#include <stdexcept>

#undef module_name
#define module_name "PFTABLE"

PrefixTable::PrefixTable(Partition p) : m_Partition(p), m_RECount(0), m_RRCount(0), m_BCalls(0), m_TotalLocs(0)
{
	Timer tmr;
	float t1, t2;
	tmr.ST();
	RefTable = new PrefixRefRelay*[cRefTableLen];
	t1=tmr.ET(); tmr.ST();
	memset(RefTable, 0, sizeof(PrefixRefRelay*)*cRefTableLen);
	t2=tmr.ET();

	Log.Message("Generating Prefix Table (pre times = %.2fs %.2fs)", t1, t2);
	Generate();
	Log.Message("Nodes in table: Relay: %i Entry: %i (%i build calls)", m_RRCount, m_RECount, m_BCalls);

	//Compress();
}
PrefixTable::~PrefixTable()
{
	Log.Verbose("Clearing prefix table");
	Clear();
	delete[] RefTable;
	Log.Verbose("Cleanup done");
}


int lastSeqTotal = 0;

void PrefixTable::Generate()
{
	for (Partition::iterator itr = m_Partition.begin(); itr != m_Partition.end(); ++itr)
	{
		m_CurGenSeq = itr->SeqId();
		CS::PrefixIteration(itr->Sequence(), itr->Length(), &PrefixTable::BuildPrefixTable, this, (Config.Exists("ref_skip") ? Config.GetInt("ref_skip", 0, -1) : 0), itr->Start());
		//CS::PrefixIteration(itr->Sequence(), itr->Length(), &PrefixTable::BuildPrefixTable, this, 0, itr->Start());
	}
}

void PrefixTable::BuildPrefixTable(ulong prefix, int pos, void* data)
{
	NGM.Stats->CurrentBase = pos;
	NGM.Stats->CurrentTotal = lastSeqTotal + pos;

	PrefixTable * _this = (PrefixTable*)data;
	_this->m_BCalls++;

	SequenceLocation tmp = {pos, _this->m_CurGenSeq};

	_this->SaveToRefTable(prefix, tmp);
}
PrefixRefRelay * PrefixTable::FindOrInsert(ulong prefix, bool insert)
{
	int prefixHash = Hash(prefix);
	PrefixRefRelay * cur = RefTable[prefixHash];

	while (cur != 0 && cur->Prefix != prefix)
		cur = cur->Next;

	if (insert && (cur == 0))
	{
		cur = new PrefixRefRelay();
		++m_RRCount;
		cur->Prefix = prefix;
		cur->Next = RefTable[prefixHash];
		RefTable[prefixHash] = cur;
	}

	return cur;
}

void PrefixTable::SaveToRefTable(ulong prefix, SequenceLocation loc)
{
	PrefixRefRelay * relay = FindOrInsert(prefix, true);

	if (relay->Entry == 0)
	{
		relay->Entry = new RefEntry();
		++m_RECount;
		NGM.Stats->UniquePrefixCount++;
		NGM.Stats->RefEntryCount++;
	}
	else if (relay->Entry->refCount >= RefEntry::MaxRefsPerEntry)
	{
		RefEntry * tmp = relay->Entry;
		++m_RECount;
		relay->Entry = new RefEntry();
		relay->Entry->nextEntry = tmp;
		relay->Entry->refTotal = tmp->refTotal;
		NGM.Stats->RefEntryCount++;
	}

	relay->Entry->ref[relay->Entry->refCount++] = loc;
	++relay->Entry->refTotal;
	++m_TotalLocs;
}
RefEntry const * PrefixTable::GetRefEntry(ulong prefix) const
{
	PrefixRefRelay const * relay = const_cast<PrefixTable*>(this)->FindOrInsert(prefix, false);
	return (relay != 0) ? relay->Entry : 0;
}

void PrefixTable::Clear()
{
	for (uint i = 0; i < cRefTableLen; ++i)
	{
		PrefixRefRelay * cur = RefTable[i];
		while (cur != 0)
		{
			PrefixRefRelay * toDelete = cur;
			cur = cur->Next;

			delete toDelete->Entry;
			delete toDelete;
		}
		RefTable[i] = 0;
	}
}

void PrefixTable::Compress()
{
	for (uint i = 0; i < cRefTableLen; ++i)
	{
		PrefixRefRelay * cur = RefTable[i];
		while (cur != 0)
		{
			CompressNode(cur);
			cur = cur->Next;
		}
	}
}

void PrefixTable::CompressNode(PrefixRefRelay * node)
{
}
