#ifndef __PREFIXTABLE_H__
#define __PREFIXTABLE_H__

#include "IRefProvider.h"
#include "RefEntry.h"
#include "Partition.h"
#include "Types.h"
#include "CSCache.h"

class PrefixTable : public virtual IRefProvider
{
public:
	PrefixTable( Partition p );
	virtual ~PrefixTable();

	//---section IRefProvider
	int GetPageCount() const { return 1; }
	RefEntry const * GetRefEntry(ulong prefix) const;
	void LoadPage(int const page) const { return; }
	//---section IRefProvider

	//---section Flattening
	static PrefixTable * GenerateFromCache(CSCache const * cache, int const page);
	void SaveToCache(CSCache * cache, int const page);
	//---section Flattening
private:
	static const uint cRefTableLen = 9999991;

	Partition m_Partition;
	PrefixRefRelay ** RefTable;
	int m_CurGenSeq;
	int m_RECount;
	int m_RRCount;
	int m_BCalls;
	ulong m_TotalLocs;

	void Generate();
	static void BuildPrefixTable(ulong prefix, int pos, void* data);
	void SaveToRefTable(ulong prefix, SequenceLocation loc);
	void Compress();
	void CompressNode(PrefixRefRelay * node);
	void Clear();

	static inline uint Hash(ulong prefix)
	{
		return prefix % cRefTableLen;
	}

	PrefixRefRelay * FindOrInsert(ulong prefix, bool insert);

	//---section Flattening
	static PrefixTable * GenerateFromBuffer(char const * pBuffer, ulong const length);

	PrefixTable();
	char const * Flatten(ulong & resultSize) const;
	char const * Flatten(char * pBuffer, ulong const bufferSize, ulong & resultSize) const;

	ulong CalcSize() const;
	static bool CheckHeader(char const * const buffer);

	friend class CSCache;
	//---section Flattening
};

CSCache const * GenerateTables(uint const threadcount, PartitionList partitions, CSCache * cache);

#endif
