#include "NGMThreads.h"
#include "Types.h"
#include "Partition.h"
#include "PrefixTable.h"
#include "Log.h"
#include "Timing.h"
#include "PlatformSpecifics.h"
#include "NGM.h"

#ifdef _WIN32
#include <conio.h>
#endif

#undef module_name
#define module_name "PGEN"

namespace PrefixTableGen
{
	NGMMutex * _mutex;
	PartitionList _partitions;
	size_t curPartition = 0;
	volatile int threadsRunning = 0;

	bool GetPartition(Partition & partition, int & pNum)
	{
		bool found;
		NGMLock(_mutex);
		if (curPartition < _partitions.size())
		{
			partition = _partitions[curPartition];
			pNum = curPartition;
			++curPartition;
			found = true;
		}
		else
		{
			found = false;
		}
		NGMUnlock(_mutex);
		return found;
	}

	NGMTHREADFUNC PrefixGenThreadFunc(void * param)
	{
		Log.Message("GenThread Start");
		CSCache * cache = (CSCache*)param;
		Partition part;
		int pNum = -1;
		Timer tmr;
		int partitions_created = 0;

		while (GetPartition(part, pNum))
		{
			tmr.ST();
			Log.Message("Generating table for partition %i", pNum);
			PrefixTable * table = new PrefixTable(part);

			table->SaveToCache(cache, pNum);
			delete table;
			Log.Message("Partition %i generated (elapsed: %.2fs)", pNum, tmr.ET());
			++partitions_created;
		}

		AtomicDec(&threadsRunning);
		Log.Message("GenThread done");
		return 0;
	}
};

using namespace PrefixTableGen;

CSCache const * GenerateTables(uint const threadcount, PartitionList partitions, CSCache * cache)
{
	Log.Message("Generating %i prefix tables on %i threads", partitions.size(), threadcount);
	_mutex = new NGMMutex();
	NGMInitMutex(_mutex);

	_partitions = partitions;

	NGMLock(_mutex);
	NGMThread * threads = new NGMThread[threadcount];
	for (uint i = 0; i < threadcount; ++i)
	{
		threads[i] = NGMCreateThread(PrefixGenThreadFunc, cache);
		AtomicInc(&threadsRunning);
	}
	NGMUnlock(_mutex);

	while (threadsRunning > 0)
	{
		while (_kbhit())
		{
			char ch = _getch();
			if ( ch == 'q')
				NGM.InitQuit();
		}
		Sleep(50);
	}
	NGMJoinAll(threads, threadcount);

	return cache;
}
