#include "NGM.h"
#include "ReadBuffer.h"
#include "PlatformSpecifics.h"
#include <sstream>
#include <stdio.h>

#undef module_name
#define module_name "RBUF"

ReadBuffer::ReadBuffer(int firstRead, int totalReads, bool generateFile) : m_File(0), m_ScoresPerRead(10), m_BytesPerRead(sizeof(LocationScore) * m_ScoresPerRead),
  m_ReadIdOffset( firstRead ), m_TotalReads( totalReads ), m_Running( false ), m_Writing( false )
{
	m_WriteBuffer.reserve(m_TotalReads);
	/*NGMInitMutex(&m_Mutex);

	if (generateFile)
	{
		m_Filename = GenerateFilename();

		m_File = fopen(m_Filename.c_str(), "w+b");
		if (m_File == 0)
		{
			Log.Error("Unable to open read buffer file %s", m_Filename.c_str());
			Fatal();
		}
		else
		{
			Log.Message("Buffering reads to %s (%i)", m_Filename.c_str(), m_File);
			InitFile();
		}
	}

	NGMCreateThread(&WriteFunc, this);

	while (!m_Running)
		;*/
}

NGMTHREADFUNC ReadBuffer::WriteFunc(void* tthis)
{
	ReadBuffer * tThis = (ReadBuffer*)tthis;
	tThis->m_Running = true;

	while (tThis->m_Running)
	{
		if (tThis->m_WriteBuffer.size() > 0)
			tThis->DumpBufferToFile();
		else
			Sleep(100);
	}

	return 0;
}

void ReadBuffer::WaitForWrite()
{
	while (m_Writing)
		Sleep(10);
}

void ReadBuffer::DumpBufferToFile()
{
	NGMLock(&m_Mutex);
	TWriteBuffer toWrite = TWriteBuffer( m_WriteBuffer );
	m_WriteBuffer.clear();

	WaitForWrite();

	m_Writing = true;
	NGMUnlock(&m_Mutex);

	Log.Green("Dumping %i", toWrite.size());
	for (TWriteBuffer::iterator itr = toWrite.begin(); itr != toWrite.end(); ++itr)
	{
		DoWrite(*itr);
	}
	m_Writing = false;
	Log.Green("Dump done");
}

ReadBuffer * ReadBuffer::ContinueFrom(std::string filename, int firstRead, int totalReads, int & partition)
{
	Log.Error("Continue_from not active");
	Fatal();

	ReadBuffer * buffer = new ReadBuffer(firstRead, totalReads, false);
	buffer->m_Filename = filename;

	buffer->m_File = fopen(buffer->m_Filename.c_str(), "r+b");
	if (buffer->m_File == 0)
	{
		Log.Error("Unable to open read buffer file %s", filename.c_str());
		Fatal();
	}

	int ret = fseek(buffer->m_File, (long)buffer->m_TotalReads * (long)buffer->m_BytesPerRead, SEEK_SET);
	if (ret != 0)
	   Log.Error("Error continuing from read buffer file (fseek error %i)", ret);
	int n = -1;
	ret = fread(&n, sizeof(int), 1, buffer->m_File);
	if (ret != 1)
	   Log.Error("Error reading last partition from read buffer file");

	partition = n;
	return buffer;
}

ReadBuffer * ReadBuffer::GenerateNew(int firstRead, int totalReads)
{
	return new ReadBuffer(firstRead, totalReads, true);
}

ReadBuffer::~ReadBuffer()
{
	m_Running = false;
	//NGMJoin(&m_Thread);

  if (m_File)
    fclose(m_File);
}

std::string ReadBuffer::GenerateFilename()
{
	std::stringstream fnss;
	char const * tmp = Config.GetString("tmp_dir");
	if (tmp == 0 || strlen(tmp) <= 0)
	{
	tmp = "/tmp";
	}
	fnss << tmp << "/ngm-buffer-" << GetPID() << ".tmp";
	return fnss.str();
}

void ReadBuffer::InitFile()
{
	int ret = fseek(m_File, (long)m_TotalReads * (long)m_BytesPerRead, SEEK_SET);
	if (ret != 0)
	   Log.Error("Error generating read buffer file (fseek error %i)", ret);
	int n = -1;
	ret = fwrite(&n, sizeof(int), 1, m_File);
	if (ret != 1)
	   Log.Error("Error generating read buffer file (fwrite)");
}

void ReadBuffer::CheckError(int result, int expected) const
{
	if (result != expected)
	{
		Log.Error("File operation returned error. This will likely lead to corrupt output data. Error = %i", ferror(m_File));
		perror(0);
	}
}

ulong lastMsg = 0;
void PrintScores(MappedRead const * read);

void ReadBuffer::Save(MappedRead const * read)
{
	Log.Verbose("Saving Read %i", read->ReadId);
	m_WriteBuffer[read->ReadId - m_ReadIdOffset] = read;

	/*	NGMLock(&m_Mutex);
	m_WriteBuffer.push_back(read);

	if (lastMsg < GetTickCount() - 1000)
	{
		Log.Green("sz: %i - c: %i", m_WriteBuffer.size(), m_WriteBuffer.capacity());
		lastMsg = GetTickCount();
	}
	NGMUnlock(&m_Mutex);*/
}

void ReadBuffer::DoWrite(MappedRead const * read)
{
	  CheckError(fseek(m_File, (long)(read->ReadId - m_ReadIdOffset) * (long)m_BytesPerRead, SEEK_SET));
	  uint count = read->Scores.size();

	  for (uint i = 0; i < count; ++i)
	    CheckError(fwrite(read->Scores[i], sizeof(LocationScore), 1, m_File), 1);

	  delete read;
}

MappedRead * ReadBuffer::Load(int const readid) const
{
	MappedRead* read = (MappedRead*)m_WriteBuffer[readid-m_ReadIdOffset];
	read->Calculated = read->Scores.size();
	Log.Verbose("loaded read %i (scores %i, calc %i)", read->ReadId, read->Scores.size(), read->Calculated);
	return read;
/*  static LocationScore * _tmp = new LocationScore[m_ScoresPerRead];

  CheckError(fseek(m_File, (long)(readid - m_ReadIdOffset) * (long)m_BytesPerRead, SEEK_SET));
  CheckError(fread(_tmp, sizeof(LocationScore), m_ScoresPerRead, m_File), m_ScoresPerRead);

  MappedRead * read = new MappedRead(readid);
  int count = 0;
  for (int i = 0; i < m_ScoresPerRead; ++i)
  {
    if (_tmp[i].Score.f > 0.0)
    {
      read->AddScore(_tmp[i]);
      ++count;
    }
  }
  read->Calculated = count;
  return read;*/
}

void ReadBuffer::FinalizePartition(int partition)
{
/*	DumpBufferToFile();

	CheckError(fseek(m_File, (long)m_TotalReads * (long)m_BytesPerRead, SEEK_SET));
	CheckError(fwrite(&partition, sizeof(int), 1, m_File), 1);

	fflush(m_File);*/
}
