#ifndef __READBUFFER_H__
#define __READBUFFER_H__

#include "MappedRead.h"
#include <string>
#include <vector>
#include "NGMThreads.h"

class ReadBuffer
{
private:
	std::string m_Filename;
	FILE * m_File;
	int const m_ScoresPerRead;
	int const m_BytesPerRead;
	int const m_ReadIdOffset;
	int const m_TotalReads;

	typedef std::vector<MappedRead const*> TWriteBuffer;

	TWriteBuffer m_WriteBuffer;

	NGMMutex m_Mutex;
	NGMThread m_Thread;
	static NGMTHREADFUNC WriteFunc(void*);
	volatile bool m_Running;
	volatile bool m_Writing;

	void WaitForWrite();
	void DumpBufferToFile();
	void DoWrite(MappedRead const * read);
	void CheckError(int result, int expected = 0) const;
	static std::string GenerateFilename();
	void InitFile();
	ReadBuffer(int firstRead, int totalReads, bool generateFile);
public:
	~ReadBuffer();

	void Save(MappedRead const * read);
	MappedRead * Load(int const readid) const;

	void FinalizePartition(int partition);

	static ReadBuffer * ContinueFrom(std::string filename, int firstRead, int totalReads, int & partition);
	static ReadBuffer * GenerateNew(int firstRead, int totalReads);
};

#endif
