#include "MappedRead.h"
#include "NGM.h"

using NGMNames::ReadStatus;

MappedRead * GenerateSingleRead(int const readid)
{
	MappedRead * read = 0;

	if (SequenceProvider.IsValid(readid))
		read = new MappedRead(readid);

	return read;
}

// Sequential (important for pairs!) read generation
MappedRead * _NGM::GenerateRead(int const readid)
{
	static MappedRead * nextRead = 0;

	if (NGM.Paired())
	{
		MappedRead * read = 0;
		if ( !(readid & 0x1) ) // Requested first read of a pair (even-numbered read id)
		{
			// Generate whole pair
			read = GenerateSingleRead(readid);
			nextRead = GenerateSingleRead(readid + 1);

			if (read != 0)
			{
				read->Paired = nextRead;
				if (nextRead == 0)
					read->SetFlag(NGMNames::NoSrcPair);
			}

			if (nextRead != 0)
			{
				nextRead->Paired = read;
				if (read == 0)
					nextRead->SetFlag(NGMNames::NoSrcPair);
			}
		}
		else
		{
			read = nextRead;
			nextRead = 0;

			if ( (read != 0) && (read->ReadId != readid) )
				Log.Error("Non-sequential read generation (req %i, got %i)", readid, read->ReadId);
		}

		return read;
	}
	else
	{
		return GenerateSingleRead(readid);
	}
}

void _NGM::DisposeRead(MappedRead * read)
{
	if (NGM.Paired() && read->Paired != 0)	// Program runs in paired mode and pair exists
	{
		if (read->Paired->HasFlag( NGMNames::DeletionPending )) // Paired read was marked for deletion
		{
			// delete paired and read
			delete read->Paired;
			delete read;
		}
		else
		{
			// Paired read is still in use...mark read for deletion
			read->SetFlag( NGMNames::DeletionPending );
		}
	}
	else
	{
		// Single mode or no existing pair
		delete read;
	}
}