#ifndef __SEQUENCEPROVIDER_H__
#define __SEQUENCEPROVIDER_H__

#include <string>

#include "Types.h"

/*
 *  SequenceProvider: Schnittstelle zu den Sequenz-Daten
*/
class _SequenceProvider
{
public:
	static _SequenceProvider & Instance();
	static void Cleanup();

	virtual void Init();

	virtual char const * const GetRefSequence(int n) const;
	virtual char const * const GetQrySequence(int n) const;

	// Gets the length of read/reference string n
	virtual int GetRefLen(int n) const;
	virtual int GetQryLen(int n) const;

	virtual int GetReadMaxLen() const;

	virtual int GetRefCount() const;
	virtual int GetQryCount() const;

	virtual char const * GetRefName(int n, int & len) const;
	virtual char const * GetQryName(int n, int & len) const;
	virtual char const * GetQualityString(int n, int & len) const;

	virtual bool IsValid(int n) const;
	virtual bool HasPair(int n) const;

	virtual void PagingUpdate();
private:
	_SequenceProvider();
	~_SequenceProvider();

	bool CheckQryNr(int n) const;
	bool CheckRefNr(int n) const;

	static _SequenceProvider * pInstance;

	struct TIdxHeader
	{
		int Signature;
		char Zero;
		char VersionMajor;
		char VersionMinor;
		char VersionStep;
		uint HeaderLength;	// Header Len
		uint EntryLength;
		uint SeqCount;	// # Sequences in File
		uint SeqFieldWidth; // Max Width (+padding) of a read, for refs 0
		uint Flags;
		uint Checksum;
	};

	struct TRefIdx
	{
		uint SeqId;
		uint Flags;
		ulong SeqStart;
		ulong DescStart;
		uint SeqLen;
		uint DescLen;
	};

	struct TQryIdx
	{
		uint SeqId;
		uint Flags;
		ulong DescStart;
		ulong QualityStart;
		uint DescLen;
		uint QualityLen;
	};

	// Files
	std::string refIdxName;
	std::string qryIdxName;
	std::string prdIdxName;

	std::string refFileName;
	std::string qryFileName;
	std::string refCplFileName;
	std::string prdFileName;

	std::string refBaseFileName;
	std::string qryBaseFileName;
	std::string prdBaseFileName;

	// The actual data, layed out flat in memory
	char const * pQryData;
	char const * pRefData;
	char const * pRefCplData;
	char const * pPrdData;

	char const * pRefBaseData;
	char const * pQryBaseData;
	char const * pPrdBaseData;

	char const * pQryIdxData;
	char const * pRefIdxData;
	char const * pPrdIdxData;
	TIdxHeader * refHeader;
	TIdxHeader * qryHeader;
	TIdxHeader * prdHeader;
	TRefIdx * refIdx;
	TQryIdx * qryIdx;
	TQryIdx * prdIdx;

	// Length of mapped files
	long qryFileLen;
	long refFileLen;
	long refCplFileLen;
	long prdFileLen;
	long refBaseFileLen;
	long qryBaseFileLen;
	long prdBaseFileLen;

	// Handles to the mapped files
	int refIdxMap;
	int qryIdxMap;
	int prdIdxMap;
	int qryMap;
	int refMap;
	int prdMap;
	int refcplMap;

	int refBaseMap;
	int qryBaseMap;
	int prdBaseMap;

	int qryMaxLen;

	int refCount;
	mutable int qryCount;

	void PrepareInputFile(std::string);
};

#define SequenceProvider _SequenceProvider::Instance()

#endif
